/*
 * WeightFunc.hpp  --  Part of the CinePaint plug-in "Bracketing_to_HDR"
 *
 * Copyright (c) 2005-2006  Hartmut Sbosny  <hartmut.sbosny@gmx.de>
 *
 * LICENSE:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/**
  @file  WeightFunc.hpp  -  Wichtungsfunktionen. 
  
  Contents:
   - WeightFuncBase        --  lowest base class (non-template)
   - WeightFunc<>          --  base template
   - WeightFunc_Triangle<> --  template of triangle functions (unused)
   - WeightFunc_U8         --  base class for uint8 weight functions
   - WeightFunc_U16        --  base class for uint16 weight functions
  
  FRAGE: Brauchen wir ueberhaupt den Template-Parameter `Unsign'? Kommt zum
   Einsatz nur im Ctor als Typ des `zmax'-Arguments und in den Op-Funktionen
      Tweight operator() (Unsign z) {...}
      Tweight operator[] (Unsign z) {...}   // zur Zeit noch int-Argument!
   Nur letzteres echter Grund. Alternative waere, ein festes `int' oder 
    `unsigned int'-Argument. Bietet das Unsign irgendeinen Vorteil? Unsere
    TNT::Array1D-Tabelle nimmt int als Argumet der []-Operatoren, dh. hier
    findet sowieso Konvertierung Unsign->int statt. Aber bei Typmischmasch
    wird gewarnt, vor allem wichtig fuer die Array-Aufrufe `[i] mit einem 
    Typ(i) > T.
   
  Noch unentschieden bin ich in der Frage, ob die Funktionen Werte \in [0..1]
   liefern sollen (meine Erwartung beim Name "Gewicht") oder, wie
   gegenwaertig im Deberece-Verfahren der Fall, Werte \in [0..zmax]. Durch
   letzteres kommt eine zusaetzlche Abhaengigkeit vom Datentyp und der Bitzahl
   hinein. Wenn nicht fuer die Balance des Gleichungsstems wichtig (was ich 
   jetzt nicht ueberblicke, aber bezweifle), sollte das vermieden werden.
   Inzwischen umgestellt auf Werte \in [0..1]
   
  SPEZIALISIERUNG: Wegen der Abstammung WeightFunc_Triangle<> --> WeightFunc<>
   gilt auch fuer die Spezialisierungen: 
      WeightFunc_Triangle<uint8> --> WeightFunc_Triangle<uint8>
   etc. Benutzen wir hingegen eine echte U8-Ableitung,
      class WeightFunc_U8 : public WeightFunc<uint8> {...}
   um U8-spezifischen Ergaenzungen anzubringen (Tabelleninit. im Ctor, Array-
   Syntax), ist eine solche U8-Klasse nicht mehr die Basisklasse von 
   WeightFunc_Triangle<uint8>. Die Form-Funktionen (Triangle etc.) muss dann fuer
   jeden Unsign-Typ extra implementiert werden.
*/
#ifndef WeightFunc_hpp
#define WeightFunc_hpp

#include "TNT/tnt_array1d.hpp"
#include "br_types.hpp"         // uint8

#include <iostream>
#include <cmath>                // sin(), M_PI


/**
  Data type of the weight values...
   
  Welchen Typ soll `Tweight' haben?
   Massiv gebraucht wird Wichtungstabelle in merge_HDR(). Dort wird zunaechst
   in double summiert und erst dann auf float (HDR) reduziert. Legte double
   nahe, float ginge aber auch, wuerde eben zuvor in double konvertiert. Ist
   noch die Platzfrage, die bei 256 Werten unkritisch, bei 65536 Werten 
   (16 Bit) aber bedenkenswert (256 KB vs. 512 KB).
*/
typedef double Tweight;  


/**===========================================================================
  
  @class WeightFuncBase 
   
  Basisklasse (kein Template) der Wichtungstemplates. Enthaelt das nicht daten-
   typabhaengige. Theoretisch koennten auch `minval, `maxval und `what noch
   hierher.
  
=============================================================================*/
class WeightFuncBase 
{
public:
    enum Shape 
    {   IDENTITY,
        IDENTITY_0,
        TRIANGLE,
        SINUS,
        UNKOWN
    };

private:
    Shape shape_;
    
protected:
    Tweight  minval_;                   // maximal weight value
    Tweight  maxval_;                   // minimal weight value
    std::string  what_;                 // description string
        
public:    
    WeightFuncBase (Shape s) 
      : shape_(s), 
        minval_(0.),
        maxval_(0.),
        what_(__func__)
      {}
    Shape shape() const                  {return shape_;}  
    Tweight minval() const      {return minval_;}   // minimal weight value
    Tweight maxval() const      {return maxval_;}   // maximal weight value
    const char* what() const    {return what_.c_str();}
};


/**===========================================================================
  
  @class WeightFunc  -  template
   
  Basistemplate der Wichtungstemplates mit (virt.) Funktionssyntax. Erben haben
   diese virtuelle Funktion zu definieren. Bereitgestellt ist dann zusaetzlich 
   die Moeglichkeit, eine Tabelle mit diesen Werten zu initialisieren und via 
   []-Op auszulesen, was schneller als der Funktionsaufruf. Da fuer Unsign==U16
   die Tabelle 65535 double- oder float-Werte umfasst, kann dieser Speicher bei
   zeitkritischen Aktivitaeten genutzt, bei speicherkritischen aber auch gespart
   werden.
   
  Die Funktionssyntax also immer moeglich, der []-Op nur nach init_table(). Bei
   U8 koennte Tabelle stets schon im Ctor angelegt werden. 
   
  Zur Verwaltung der Tabelle nutzen wir die referenzzaehlende Haendelklasse 
   TNT::Array1D<> (Element, keine Ableitung).  
    
  NOTE: Selbst wenn die virt. Funktion auch im Basistemplate schon definiert
   und nicht abstrakt gelassen wuerde, um Instanzen auch von Weight_U8 etc. zu
   ermoeglichen, koennten wir trotzdem keine *Objekte* als allg. Weight-Variablen
   benutzen, d.h. Dinge machen wie:  
     
     class A { 
        Weight_U8  w_;                          // allg. Variable
        set_weight (Weight_U8 & w)  {w_=w;}     // kopiere `w auf `w_
     };
     A a;
     Weight_U8_Triangle  tri;
     a.set_weight (tri);        // Kopiert `triangle auf `a.w_?  FEHLER!
     
   Denn beim Kopieren auf den Vorfahr wuerde zwar die Tabelle (da Element)
   kopiert, die virt. Funktion aber bliebe beim Zielobjekt unveraendert. Die
   einzige Moeglichkeit zur Haltung einer allg. Variable sind *Zeiger*.
   
     class A {
        Weight_U8* p_;                          // Basis-Zeiger
        set_pweight (Weight_U8* p)  {p_=p;}
     };
     A a;
     Weight_U8_Triangle  tri;
     a.set_weight (&tri);
   
   Bei der Nutzung von `p_ als `(*p_)(z) wird die richtige virt. Funktion
   herausgesucht.
   
=============================================================================*/
template <typename Unsign>
class WeightFunc : public WeightFuncBase
{
protected:
    
    TNT::Array1D<Tweight>  table_;
    Unsign  zmax_;                      // maximal z-argument

public:
    
    WeightFunc (Shape s, Unsign zmax)   // Ctor; `table_ as Null-Array
      : WeightFuncBase (s),
        zmax_(zmax)
      {
        what_=__func__;
      }    
    //  virt. Dtor, um die Warnungen zu vermeiden
    virtual ~WeightFunc()       {/*std::cout << "Dtor WeightFunc '" << what_ << "'\n";*/}
      
    void init_table();     
    void remove_table()         {table_ = TNT::Array1D<Tweight>();}  // Null-Array
    bool have_table() const     {return !table_.is_null();}
    int  table_dim() const      {return table_.dim();}
    
    const TNT::Array1D<Tweight> & table() const {return table_;}
    
    const Tweight & operator[] (int i) const    {return table_[i];}  // Hier kein Unsign Arg ??
          Tweight & operator[] (int i)          {return table_[i];}
    
    Unsign  zmax() const        {return zmax_;}     // maximal possible z-value
    
    //====================
    // abstract interface:
    //====================
    virtual Tweight operator() (Unsign z) const = 0;          // abstract
};


/**+***************************************************************************
  Create the table and initialize it using the virtual function.
******************************************************************************/
template <typename Unsign>
void WeightFunc<Unsign>::init_table()
{
    table_ = TNT::Array1D<Tweight> (zmax_ + 1); // kein Ueberlauf, da bei numer. Ops
                                                //  stets vorher in int gewandelt
    for (int z=0; z < table_.dim(); z++)  
      table_[z] = (*this)(z);                   // uses the virt. function
}


#if 1
/**===========================================================================
  
  @class WeightFunc_Triangle  -  template   (unused, exercise)
  
  Beispiel eines von WeightFunc<> abgeleiteten Dreieck-TEMPLATES, d.h. mit 
   einer allgemeinen (template-tauglichen) Definition der virt. Funktion.

=============================================================================*/
template <typename Unsign>
class WeightFunc_Triangle : public WeightFunc<Unsign> 
{
    Unsign zmin_, zmid_;                        // neben `zmax_ aus base class

public:
    
    WeightFunc_Triangle (Unsign zmax)           // Ctor
      : 
        WeightFunc<Unsign> (WeightFuncBase::TRIANGLE, zmax)
      { 
        zmin_   = 0;
        zmid_   = (zmin_ + zmax) / 2;
        WeightFunc<Unsign>::minval_ = 0.;
        WeightFunc<Unsign>::maxval_ = (*this) (zmid_);  // maximum at the middle
      }

    Tweight operator() (Unsign z) const         // virtual Triangle function; could
      {                                         //  be simplified by zmin==0.
        //return (z <= zmid_) ? (z - zmin_) : (zmax_ - z);
        return ((z <= zmid_) ? (z - zmin_) : (WeightFunc<Unsign>::zmax_ - z)) / Tweight(zmid_);
      }
};
#endif

//======================
// Spezialisierungen...
//======================

/**===========================================================================

  @class WeightFunc_U8            -  Base class of U8 weight objects
  @class WeightFunc_U8_Identity   -  Identity function
  @class WeightFunc_U8_Identity_0 -  Identity function with 0-borders
  @class WeightFunc_U8_Triangle   -  Triangle shape function
  @class WeightFunc_U8_Sinus      -  Sinus shape function 

*============================================================================*/
class WeightFunc_U8 : public WeightFunc <uint8>
{
public:
    WeightFunc_U8 (Shape shape) : WeightFunc<uint8> (shape, 255)  // Ctor
      {
        what_ = __func__;
        //init_table();   // For U8 as default? Aber nie hier, benuetzte 
      }                   //  falsche virt. Funktion!!
};

class WeightFunc_U8_Identity : public WeightFunc_U8
{
public:
    WeightFunc_U8_Identity() : WeightFunc_U8 (IDENTITY)  // Ctor
      {
        WeightFunc<uint8>::minval_ = 1.;
        maxval_ = 1.;
        what_   = __func__;
        //init_table();                           // for U8 as default???
      }
    
    Tweight operator() (uint8 z) const          // virtual Identity function
      {
        return 1.;
      }
};

class WeightFunc_U8_Identity_0 : public WeightFunc_U8
{
public:
    WeightFunc_U8_Identity_0() : WeightFunc_U8 (IDENTITY_0)  // Ctor
      {
        minval_ = 0.;
        maxval_ = 1.;
        what_   = __func__;
        //init_table();                           // for U8 as default???
      }
    
    Tweight operator() (uint8 z) const          // virtual Identity function
      {
        if (z==0 || z==255) return 0.0; else return 1.0;
      }
};


class WeightFunc_U8_Triangle : public WeightFunc_U8
{
public:
    WeightFunc_U8_Triangle() : WeightFunc_U8 (TRIANGLE)  // Ctor
      {
        minval_ = 0.;
        maxval_ = (*this) (127);                // maximum at the middle
        what_   = __func__;
        //init_table();                           // for U8 as default???
      }
    
    Tweight operator() (uint8 z) const          // virtual triangle function; could
      {                                         //  be simplified by zmin==0.
        //return (z <= zmid_) ? (z - zmin_) : (zmax_ - z);
        //return (z <= 127) ? (z - 0) : (255 - z);
        return ((z <= 127) ? (z - 0) : (255 - z)) / 127.;
      }
};

class WeightFunc_U8_Sinus : public WeightFunc_U8
{
    Tweight factor_;
public:
    WeightFunc_U8_Sinus()                       // Ctor
      :
        WeightFunc_U8 (SINUS)
      {
        factor_ = M_PI / 255.;
        minval_ = 0;
        maxval_ = (*this) (127);                // maximum at the middle
        what_   = __func__;
        //init_table();                           // for U8 as default???
      }
    
    Tweight operator() (uint8 z) const          // virtual Sinus function
      {
        return sin (z * factor_);               // Ergebnis '0' an den Raendern
      }                                         //  sollte garantiert werden!
};


/**===========================================================================

  @class WeightFunc_U16            -  base class of U16 weight objects
  @class WeightFunc_U16_Identity   -  Identity function
  @class WeightFunc_U16_Identity_0 -  Identity function with 0-borders
  @class WeightFunc_U16_Triangle   -  Triangle shape function
  @class WeightFunc_U16_Sinus      -  Sinus shape function

*============================================================================*/
class WeightFunc_U16 : public WeightFunc <uint16>
{
public:
    WeightFunc_U16 (Shape shape) : WeightFunc<uint16> (shape, 65535)  // Ctor
      {
        what_=__func__;
      }
};


class WeightFunc_U16_Identity : public WeightFunc_U16
{
public:
    WeightFunc_U16_Identity() : WeightFunc_U16 (IDENTITY)
      {
        minval_ = 1.;
        maxval_ = 1.;
        what_   = __func__;
      }
    
    Tweight operator() (uint16 z) const          // virtual Identity function
      {
        return 1.;
      }
};

class WeightFunc_U16_Identity_0 : public WeightFunc_U16
{
public:
    WeightFunc_U16_Identity_0() : WeightFunc_U16 (IDENTITY_0)
      {
        minval_ = 0.;
        maxval_ = 1.;
        what_   = __func__;
      }
    
    Tweight operator() (uint16 z) const          // virtual Identity function
      {
        if (z==0 || z==65535) return 0.; else return 1.0;
      }
};

class WeightFunc_U16_Triangle : public WeightFunc_U16
{
public:
    WeightFunc_U16_Triangle() :  WeightFunc_U16 (TRIANGLE)  // Ctor
      {
        minval_ = 0.;
        maxval_ = (*this) (32767);              // maximum at the middle 
        what_   = __func__;
      }
    
    Tweight operator() (uint16 z) const         // virtual Triangle function; could
      {                                         //  be simplified by zmin==0.
        //return (z <= zmid_) ? (z - zmin_) : (zmax_ - z);
        return ((z <= 32767) ? (z - 0) : (65535 - z)) / Tweight(32767);
      }
};

class WeightFunc_U16_Sinus : public WeightFunc_U16
{
    Tweight factor_;
public:
    WeightFunc_U16_Sinus() :  WeightFunc_U16 (SINUS)  // Ctor
      {
        factor_ = M_PI / 65535.;
        minval_ = 0.;
        maxval_ = (*this) (32767);              // maximum at the middle
        what_   = __func__;
      }
    
    Tweight operator() (uint16 z) const         // virtual Sinus function
      {
        return sin (z * factor_);               // Ergebnis '0' an den Raendern         
      }                                         //  sollte garantiert werden!
};

#if 0
/**===========================================================================

  @class WeightFunc_U8           -  base class of U8 weight functions
  @class WeightFunc_Triangle_U8  -  the Triangle-function for U8 data

*============================================================================*/
typedef WeightFunc          <uint8>   WeightFunc_U8;
typedef WeightFunc_Triangle <uint8>   WeightFunc_Triangle_U8;


/**===========================================================================
  
  @class WeightFunc_U16           -  base class of U16 weight functions
  @class WeightFunc_Triangle_U16  -  the Triangle-function for U16 data

*============================================================================*/
typedef WeightFunc          <uint16>  WeightFunc_U16;
typedef WeightFunc_Triangle <uint16>  WeightFunc_Triangle_U16;
#endif


/**==========================================================================\n
  new_WeightFunc_U8()  &&  new_WeightFunc_U16():
  
  Create a dynamic WeightFunc object of a given shape. 
  @Return Pointer of the object or NULL, if creation fails.
*============================================================================*/
WeightFunc_U8*  new_WeightFunc_U8  (WeightFuncBase::Shape);
WeightFunc_U16* new_WeightFunc_U16 (WeightFuncBase::Shape);

void list (const WeightFunc_U8 & w);


#endif // WeightFunc_hpp

// END OF FILE
