/*
 * Z_MatrixGenerator.hpp  --  Part of the CinePaint plug-in "Bracketing_to_HDR"
 *
 * Copyright (c) 2005-2006  Hartmut Sbosny  <hartmut.sbosny@gmx.de>
 *
 * LICENSE:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/**
  Z_MatrixGenerator.hpp
*/
#ifndef Z_MatrixGenerator_hpp
#define Z_MatrixGenerator_hpp

#include <cstdio>                  // printf() 
#include "br_types_array.hpp"      // ChannelFollowUpArray
#include "br_macros.hpp"           // ASSERT()

#undef BR_DEBUG
#include "br_macros_varia.hpp"     // SPUR()

 
/**===========================================================================
  
  @class  Z_MatrixGenerator - template
 
  Erzeugung einer "Z-Matrix" Z[i,j]<Unsign> der Dimension [nselect, nlayers]
   aus dem von `FollowUpValues_*' gelieferten Statistikfeld vom Typ
   `ChannelFollowUpArray'. Z[i,j]: z-Wert am Ort `i in Aufnahme `j, wobei 
   "Ort i" abstrakt eine Position im Statistikfeld. Die Matrix ist dann
   Ausgangspunkt der Kennlinenberechnung. Klasse umfasst im wesentlichen
   nur die Funktion `create_Z()', dennoch als Klasse angelegt, um Debug-
   Hilfsmittel (`list_InfoField()') einfach unterzubringen.
 
  Benoetigt keinen Zugriff auf Pack-Daten, daher auch keine Abhaengigkeiten
   hereingebracht. Zu konstruieren lediglich mit `z_max'-Wert. Dann
   `create_Z(..)' mit dem ChannelFollowUpArray eines Kanals aufrufen, 
   Dimensionen dieses Array2D's enthalten alle weiteren benoetigten Daten.
 
  Zum Template-Parameter `Unsign': Typ des Ctor-Arguments `z_max' und des
   returnierten Array2D (der "Z-Matrix"). `ChannelFollowUpArray' hingegen
   frei davon. `z_max' koennte leicht auch unsigned sein. Das returnierte 
   Feld koennte  verallgemeinernd float-Zahlen statt Unsign-Daten enthalten.
   Dann Template unnnoetig. Allerdings ginge in ResponseSolver dann keine
   Wichtungs*tabelle* mit Index-Syntax mehr, dann braeuchte es dort Funktion
   mit float-Argument! (In merge_() hingegen kommen nur echte z-Werte vor.)
   Global und "ohne weiteres" koennte Z-Matrix hoechstens ein unsigend-Feld
   werden.
 
  Note: Es sind verschiedene Varianten einer Z-Matrix-Generierung denkbar,
   dies ist nur eine davon.
 
*============================================================================*/
template <typename Unsign>
class Z_MatrixGenerator 
{
private:

    struct ValueInfo {bool ok; int pred; int succ; bool used;};
      // Eintrag eines [n_zvals]-Infofeld `F', darin fuer jede Zeile von `F'
      //  markiert, ob brauchbar (ok==true), welches die vorhergehende und
      //  welches die nachfolgende brauchbare Zeile, und ob Zeile bereits fuer
      //  Z-Matrix benutzt...

    Unsign z_max;

public:

    Z_MatrixGenerator (Unsign zMax) : z_max (zMax)  {}

    TNT::Array2D<Unsign> 
      create_Z (const ChannelFollowUpArray &, int refpic, int nselect);

private:
          
    void list_InfoField (const ValueInfo*, int) const;   // Debug helper
};



/**+*************************************************************************\n
  list_InfoField()
******************************************************************************/
template <typename Unsign>
void 
Z_MatrixGenerator<Unsign>::list_InfoField (const ValueInfo* aa, 
                                           int n_zvals )    const
{
    printf("List InfoField...\n");

    for (int i=0; i < int(n_zvals); i++)
      printf("%6d: ok=%d, pred=% d, succ=% d\n", 
          i, aa[i].ok, aa[i].pred, aa[i].succ);
}


/**+*************************************************************************\n
  create_Z()
 
  Erzeugung einer "Z-Matrix" der Dimension [nselect, nImages] aus einem Feld 
   gemittelter Folgewerte eines Kanals. Typ der Z-Matrix-Eintraege derzeit 
   `Unsign', `float' waere moegliche Alternative.
 
  @param F: (I) [nImages, n_zvals]-Array2D<float> gemittelter Folgewerte
      eines Kanals wie es von FollowUpValues-Klassen geliefert wird.
  @param refpic: (I) Ref-Bild bei Erstellung von `F'. Not used till now
  @param nselect: (I) Anzahl auszuwaehlender Werte (Stuetzstellen)
  @return: [nselect, nImages]-Array2D<Unsign>
 
  Von den 'n_zvals' moegl. Unsign-Werten [0...z_max] sind `nselect' auszuwaehlen. 
   Wie? Gleichabstaendig oder auch unter Beruecksichtigung ihrer Streuung oder..?
 
  Gegeben [n_zvals]-Feld `F' von Zeilen (oder Spalten) zu je `nImages'
   Elementen. Aus den `n_zvals' sind `nselect' moeglichst gleichabstaendig
   auszuwaehlen. Dabei koennen Zeilen allerdings unbrauchbar sein (n==0 oder auch
   innerhalb einer Zeile keine zwei benachtbarten Bilder im AB). Diese Zeilen
   sind auszulassen und stattdessen naechstbenachtbarte zu nehmen. Dazu dient
   unten das Infofeld `aa[]`.
 
  In der Tat sind nur Zeilen brauchbar, darin wenigstens in zwei Bildern die
   Daten im Arbeitsbereich. Wertkombinationen 0--x oder x--zmax sind wertlos,
   weil 0 *alles* Unterbelichtete aufsammelt und zmax *alles* Ueberbelichtete.
 
  FRAGE: Ist fuer `Z' der Array2D<Unsign>-Typ wirklich erforderlich (wenn
   singulaer im Programm, repliziert das einigen Code), oder reicht auch was
   Elementareres bzw. ein schon benutzter Array2D-Typ, z.B. Array2D<float>?
 
  @TODO: Die Impl. von "zwei benachtbarte im AB" unten erkennt zwar Situation
   "xx000", aber nicht "0xx00", dh wenn zwei beachtbarte spaeter. Zudem bisher
   nur fuer `first' implementiert, `last' noch gemaess "irgend zwei" ermittelt!
 
  @NOTE: Das Endresultat sind Unsign-Daten (uchars oder ushorts), doch bei
   numerischen Ops mit uchars und ushorts werden diese zunaechst immer in 
   ints konvertiert. Ist es dann nicht effektiver, bei den Unsigns `z' und
   `z_max' gleich mit ints zu rechnen?

  @NOTE: Operation mit Unsign z_max der Art `z_max + 1' sind sowieso nur bis
   ushort (uint16) moeglich. Fuer den maximalen uint32 gaebe das 0. 
******************************************************************************/
template <typename Unsign> 
TNT::Array2D<Unsign> 
Z_MatrixGenerator<Unsign>::create_Z (

        const ChannelFollowUpArray &F, 
        int refpic, 
        int nselect )
{ 
    int nlayers_ = F.dim1();
    int n_zvals  = F.dim2();    // statt `n_zvals besser `n_grid

    ASSERT((int)z_max+1 == n_zvals, RELAX);     /// @BUG bei U16 z_max+1 != n_grid!

    // Bereich feststellen, darin wenigstens zwei Werte im Arbeitsbereich
    int  first, last;             // gleicher Typ wie `n_zvals'!
    bool ok;

    ok = false;                   // noch keine zwei Werte im AB
    for (first=0; first < n_zvals; first++)
    {
      bool ok1 = false;           // noch kein Wert im AB
      for (int p=0; p < nlayers_; p++)
      {
        Unsign z = Unsign(F[p][first].z + 0.5);  // runden fuer > 0
#if 0   
        // Zwei AB-Werte in *irgendzwei* Bildern...
        if (z > 1 && z < z_max-1)     // im Arbeitsbereich
//    if (z > 0 && z < z_max)       // im Arbeitsbereich
        {   if (ok1)                  // ein Wert war schon im AB
            {   ok = true;            // folglich jetzt ein zweiter
                printf ("ok-break bei first=%i, p=%i\n", first,p);
                break;                // -> p-Schleife verlassen; wegen
            }                         // ok dann auch first-Schleife
            else
                ok1 = true;           // 1. Wert im AB
        }
#else
        // Zwei AB-Werte in *benachtbarten* Bildern...
        if (ok1)                      // ein Wert war schon im AB
        {   if (z > 1 && z < z_max-1)
            {   ok = true;            // folglich jetzt ein zweiter
                break;                // -> p-Schleife verlassen; wegen
            }                         //  ok dann auch first-Schleife
            else
                break;                // p-Schleife verlassen, da nicht
        }                             //  mehr benachbart
        else                          // noch kein Wert im AB
            ok1 = (z > 1 && z < z_max-1); // Erster Wert im AB?
#endif
      }
      if (ok) break;
    }

    ok = false;                       // noch keine zwei Werte im AB
    for (last=z_max; last > first; last--)
    {
      bool ok1=false;
      for (int p=0; p < nlayers_; p++)
      {
        Unsign z = Unsign(F[p][last].z + 0.5);   // runden fuer > 0
        if (z > 1 && z < z_max-1)     // im Arbeitsbereich (+1)
//      if (z > 0 && z < z_max)       // im Arbeitsbereich (exakt)
        {   if (ok1)                  // ein Wert war schon im AB
            {   ok = true;            // folglich jetzt ein zweiter
                break;                // --> p-Schleife verlassen; wegen
            }                         // ok dann auch last-Schleife
            else
                ok1 = true;           // 1. Wert im AB
        }
      }
      if (ok) break;
    }
    // Das ueberlappende Intervall ist also [first..last].
    printf ("%s(): first = %u, last = %u\n", __func__, first, last);
#if 0
    ValueInfo  aa [n_zvals];   // ISO C++ forbids variable-size array
#else
    ValueInfo  aa[256];         // TODO  gilt nur fuer U8!
#endif    
      // Ein [n_zvals]-Infofeld von {bool ok; int pred; int succ; bool used;},
      //  darin fuer jede Zeile von `F' markiert, ob brauchbar (ok==true),
      //  welches die vorhergehende und welches die nachfolgende brauchbare
      //  Zeile, und ob Zeile bereits fuer Z-Matrix benutzt...

    // Initialization of `aa[]'... 
    // Evaluate `aa[].ok'...
    for (int i=0;      i < first;   i++)  {aa[i].used = aa[i].ok = false;}
    for (int i=last+1; i < n_zvals; i++)  {aa[i].used = aa[i].ok = false;}
    for (int i=first;  i <= last;   i++)
    {
      aa[i].used = false;
      aa[i].ok = true;        
      for (int p=0; p < nlayers_; p++)  // ok if n != 0...
      {
        if (F[p][i].n == 0) {aa[i].ok = false;  break;}
      }
    }
    // Evaluate `pred' and `succ'...
    int pred = -1;
    for (int i=0; i < int(n_zvals); i++)
    {
      aa[i].pred = pred;
      if (aa[i].ok) pred = i;
    }
    int succ = -1;
    for (int i=n_zvals-1; i >= 0; i--)
    {
      aa[i].succ = succ;
      if (aa[i].ok) succ = i;
    }
      
    //list_InfoField (aa, n_zvals);
        
    // Number of useful lines...
    int nuseful = 0;
    for (int i=0; i < n_zvals; i++) 
      if (aa[i].ok) nuseful++;

    // How many of `nselect' are realy possible...
    int myselect;

    if (nuseful < 2)
    {
      printf ("%4d: %s(): Less than 2 useful lines! Return NULL-Array.\n", __LINE__,__func__);
      return TNT::Array2D<Unsign>();       // Null-Array
    }
    else if (nuseful < nselect)
    {
      myselect = nuseful;             // > 1 garantiert
      printf ("%s(): Only %d useful lines of %d desired\n", 
        __func__, myselect, nselect);
    }
    else
      myselect = nselect;             // > 1 garantiert

    printf("nselect = %d,  myselect = %d,  nuseful = %d,  n_zval = %d\n", 
      nselect, myselect, nuseful, n_zvals);
      
    TNT::Array2D<Unsign>  Z (myselect, nlayers_);    // allokiert

    // Jetzt ungefaehr gleichabstaendig auswaehlen...
    double dd = double(last-first) / (myselect-1);   // dd>=1 garantiert

    for (int i=0; i < myselect; i++)
    {
      size_t k = size_t(first + i*dd + 0.5);         // runden fuer > 0
      SPUR(("\twanted k = %d\n", k));
      
      // If selected line not ok, find the nearest suitable...
      if (! aa[k].ok || aa[k].used)
      {
        // Find nearest predecessor and successor not yet used...
        int pred = aa[k].pred;
        int succ = aa[k].succ;
        while (succ >= 0 && aa[succ].used) succ = aa[succ].succ; 
        while (pred >= 0 && aa[pred].used) pred = aa[pred].pred;
        if (pred < 0) 
        { 
          aa[k].pred = -1;    // Stopper fuer nachfolgende
          if (succ < 0) 
            printf("%s(): both borders reached: Should not happen!\n",__func__);
          else 
            k = succ;
        }
        else if (succ < 0)    // pred >= 0
        {
          k = pred;
        }
        else                  // pred >=0 && succ >= 0
        {
          if (k-pred < succ-k)
            k = pred;
          else
            k = succ;
        }
        SPUR(("\tused k = %d\n", k));
      }

      for (int p=0; p < nlayers_; p++) 
      {
        //if (F[p][k].n == 0) printf("i=%d p=%d: n == 0 !!\n", k,p);
        Z [i][p] = Unsign(F[p][k].z + 0.5);           // runden fuer > 0
        aa[k].used = true;
      }
    }

    //list_InfoField (aa, n_zvals);
    //::list (Z);

    return Z;
}

#endif  // Z_MatrixGenerator_hpp

// END OF FILE
