/*
 * mergeHdr_PackScheme2D_RGB.cpp  --  Part of the CinePaint plug-in "Bracketing_to_HDR"
 *
 * Copyright (c) 2005-2006  Hartmut Sbosny  <hartmut.sbosny@gmx.de>
 *
 * LICENSE:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/**
  mergeHdr_PackScheme2D_RGB.cpp
  
  Contents: 
   - merge_Hdr_RGB_U8()
   - merge_LogHdr_RGB_U8()
   - merge_Hdr_RGB_U16()
   - merge_LogHdr_RGB_U16()
*/

#include <cassert>      
#include <iostream>                  // cout
#include <limits>                    // numeric_limits<>
#include <cmath>                     // exp()
#include <fstream>                   // ofstream

#include "TNT/tnt_array1d.hpp"
#include "TNT/tnt_stopwatch.hpp"

#include "Rgb.hpp"                   // Rgb<>
#include "Rgb_utils.hpp"             // cout for Rgb<>
#include "br_PackImgScheme2D.hpp"    // PackImgScheme2D_RGB_U*
#include "br_Image.hpp"              // class Image
#include "WeightFunc.hpp"            // WeightFunc_U*
#include "ProgressInfo.hpp"
#include "mergeHdr_PackScheme2D_RGB.hpp"
#include "ResponseFunc_U16.hpp"      // ResponseFunc_U16
#include "br_ImgScheme1D.hpp"        // ImgView1D<> (only of View1D example)



namespace br {

using namespace TNT;
using std::min;
using std::max;
using std::cout;


/**  Local used bitmasks (used in BadPixelProtocol::out())  */
enum WhichBits {
    SET_NONE   =  0,
    SET_MIN_R  =  1,          ///< R set to min
    SET_MIN_G  =  2,          ///< G set to min
    SET_MIN_B  =  4,          ///< B set to min
    SET_MAX_R  =  8,          ///< R set to max
    SET_MAX_G  = 16,          ///< G set to max
    SET_MAX_B  = 32           ///< B set to max
};
const unsigned  ALL_MIN_BITS = SET_MIN_R | SET_MIN_G | SET_MIN_B;
const unsigned  ALL_MAX_BITS = SET_MAX_R | SET_MAX_G | SET_MAX_B;


/**  Name of the bad pixel protocol file */
const char* filename_bad_hdr_pixel_ = "bad_hdr_pixel.txt";

/**  Maximal number of lines (bad pixel) to output to file and stdout  */
unsigned  bad_pixel_protocol_max_lines_ = 1000;


/**  Output the bit pattern (Muster) of `arg (debug helper) */
template <typename T>
void bitout (T arg)
{
    for (int i=sizeof(T)*8-1; i>=0; i--)
      cout << bool(arg & (1 << i));
}

/**===========================================================================
  
  @class BadPixelProtocol
  
  Output a protocol of bad HDR pixels to file and/or stdout. Used by the 
   merge_HDR_*() functions.
  
=============================================================================*/
class BadPixelProtocol 
{
    std::ofstream  file_;
    bool           to_stdout_;          // output to stdout
    unsigned       lines_;              // counts the recorded lines (bad pixels)

public:
    BadPixelProtocol (const char* fname, bool to_stdout = false);
    
    void out (int x, int y, const Rgb<double>& sum_w, 
              const Rgb<float>& hdr_pixel,
              const Rgb<float>& val_min, 
              const Rgb<float>& val_max  );
              
    const std::ofstream & file() const  {return file_;}
    std::ofstream &       file()        {return file_;}
};

/**+*************************************************************************\n
  Ctor
  @param fname: Name of bad pixel file. fname==0 or "" means: no file output.
  @param to_stdout: Output to stdout? (y/n)
  
  @NOTE Default Ctor von `ofstream gibt etwas != 0, ein file.open() danach
   einen Null-Stream (doppelt geoeffnet?). Um einen Null-Stream zu erzeugen, muss
   mit file(0) oder file("") konstruiert werden.
******************************************************************************/
BadPixelProtocol::BadPixelProtocol (const char* fname, bool to_stdout)
  : file_ (0),
    to_stdout_ (to_stdout),
    lines_ (0)
{ 
    if (fname && *fname) {      // fname==0 or fname=="" means: no file output
      file_.open (fname);
      if (file_)
        file_ << "# List of bad HDR pixel. Generated by CinePaint plug-in 'bracketing_to_hdr'.\n"; 
      else
        std::cerr << "Br2Hdr-WARNING **: Could not open bad pixel protocol file \""<< fname <<"\".\n";
    }        
}

/**+*************************************************************************\n
  out()  -  output a line to `file_ and/or to stdout for a bad pixel.
  @param x,y: pixel coordinates
  @param sum_w: weight of that pixel. sum_w==0 is criteron for a "bad pixel"
  @param h: the HDR value
  @param val_min,val_max: minimal and maximal HDR value in that image series, 
     as it is used in the merge_HDR() function.
******************************************************************************/
void BadPixelProtocol::out (int x, int y, const Rgb<double>& sum_w, 
                            const Rgb<float>& h, 
                            const Rgb<float>& val_min, 
                            const Rgb<float>& val_max)
{
    if (lines_ > bad_pixel_protocol_max_lines_)     // recording limit reached
      return;
    
    unsigned  which = 0;
      
    if (sum_w.r == 0) 
      if (h.r == val_min.r)  which |= SET_MIN_R;
      else                   which |= SET_MAX_R;
    
    if (sum_w.g == 0) 
      if (h.g == val_min.g)  which |= SET_MIN_G;
      else                   which |= SET_MAX_G;
    
    if (sum_w.b == 0) 
      if (h.b == val_min.b)  which |= SET_MIN_B;
      else                   which |= SET_MAX_B;
        
    if (!which) return;         // no bad pixel
      
    if (lines_ == bad_pixel_protocol_max_lines_) {
      file_ << "More than " << lines_ << " bad pixels. Recording abborted.\n";
      if (to_stdout_) cout << "More than " << lines_ << " bad pixels. Recording abborted.\n";
      ++ lines_;
      return;
    }
    
    if (which & ALL_MIN_BITS)
      if (which == ALL_MIN_BITS) 
      {   file_  << "pure MIN ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';
          if (to_stdout_)
            cout << "pure MIN ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';
      }  
      else 
      {   file_  << "part MIN ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';           
          if (to_stdout_) 
            cout << "part MIN ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';           
      }      
        
    if (which & ALL_MAX_BITS)
      if (which == ALL_MAX_BITS) 
      {   file_  << "pure MIN ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';
          if (to_stdout_)
            cout << "pure MIN ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';
      }
      else 
      {   file_  << "part MAX ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';           
          if (to_stdout_)
            cout << "part MAX ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';           
      }
      
    ++ lines_;
}


/**+*************************************************************************\n
  merge_Hdr_RGB_U8()  --  description: see ~.hpp
  
  @NOTE Wenn die z-Werte aller Bilder (eines Kanals) an einem Pixel das Gewicht 0 erhielten,
   kann die Operation "sum_e / sum_w" nicht ausgefuehrt werden, der HDR-Wert (dieses Kanals)
   ist an dieser Stelle unbestimmt. Wir arbeiten unten mit der *Unterstellung*,
   dass Wichtungsfunktionen w(z) nur an den Raendern und in zusammenhaengenden Bereichen
   Null liefern koennen, d.h. es gebe zwei Schwellwerte `z_threshold_min und 
   `z_threshold_max mit
      
      w(z) == 0 fuer  z .<. z_threshold_min (zu dunkel -> min. HDR-Wert)\n
      w(z) == 0 fuer  z .>. z_threshold_max (zu hell -> max. HDR-Wert)\n
      w(z) != 0 sonst.
   
   Stossen wir auf ein sum_w == 0, wird anhand des z-Wertes entschieden, ob der 
   HDR-Wert auf max. oder min. HDR-Wert gesetzt wird. Fuer Wichtungsfunktionen mit
   Nullbereichen in der Mitte wuerde das nicht mehr funktionieren, der zweite else-
   Zweig unten verwiese nicht mehr notwendig auf z.>.z_threshold_max.

  @NOTE Die Klaerung, ob null-gewichtete HDR-Werte auf Min- oder Max-HDR-Wert zu 
   setzen sind anhand der Abfrage 
      
      if (sum_w.r != 0)  berechne sum_e.r/sum_w.r;   \n
      else if (pixel.r .<. z_threshold_min) set_min; \n
      else                                  set_max;
   
   funktioniert auch fuer Wichtungsfunktionen, die gar nicht oder nur einseitig 
   Null werden: Wenn gar nicht Null, kann sum_w==0 nicht auftreten, wenn einseitig
   links Null, kann sum_w nur fuer linke z.<.z_threshold_min Null werden und das 
   ist der erste else-Zweig, wenn einseitig rechts Null, ist z_threshold_min==0
   und die Bedingung z.<.z_threshold_min kann fuer solche z nicht zutreffen, 
   sprich, es wird richtigerweise im zweiten else-Zweig auf Max gesetzt.
   
  @NOTE Das Markieren abgeschnittener (null-gewichteter) HDR-Werte durch 
   Kontrastfarben erfolgt wie folgt: Lag ein HDR-Wert am unteren Rand (wuerde
   auf Min-HDR gesetzt), bekommt er den Wert 1.0 (Kontrast zur angenommenen
   dunklen Umgebung), lag er am oberen Rand (wuerde auf Max-HDR gesetzt), bekommt
   er den Wert 0.0 (Kontrast zur angenommenen hellen Umgebung. Schief abgeschnittene
   HDR-Pixel erscheinen dadurch farbig:
    unten: nur R abgeschnitten --> RGB = (1, wenig, wenig)  (rot)
           nur G abgeschnitten --> RGB = (wenig, 1, wenig)  (gruen)
           R+G   abgeschnitten --> RGB = (1, 1, wenig)
    oben:  nur R abgeschnitten --> RGB = (0, viel, viel)
           R+G   abgeschnitten --> RGB = (0, 0, viel)       (blau)
           
  @TODO Es koennte die Option angeboten werden, schief abgeschnittene Pixel (die
   naemlich auffallen) auf schwarzen Min-HDR- bzw. weissen Max-HDR-Wert zu setzen.
   Oder anhand der Umgebung korrigieren? Welches natuerlich nur moeglich, wenn 
   die Umgebung nicht selbst gestoert.
******************************************************************************/
br::Image    // IMAGE_RGB_F32
merge_Hdr_RGB_U8 (const PackImgScheme2D_RGB_U8 & pack,
                  const Array1D<double> & logX_R,
                  const Array1D<double> & logX_G,
                  const Array1D<double> & logX_B,
                  const WeightFunc_U8   & weight,
                  bool                  mark_bad_pixel, 
                  ProgressInfo*         progressinfo,
                  bool                  protocol_to_file,
                  bool                  protocol_to_stdout  )
{
    cout <<__func__<< "()...\n";
    assert (logX_R.dim() == 256);
    assert (logX_G.dim() == 256);
    assert (logX_B.dim() == 256);
    assert (weight.have_table());
    //assert (weight.table_dim() == 256);

    //  Evaluate protocol stuff
    const char*       fname = protocol_to_file ? filename_bad_hdr_pixel_ : 0;
    BadPixelProtocol  protocol (fname, protocol_to_stdout);
    bool              make_protocol = protocol.file() || protocol_to_stdout;
    
    cout << "\t(to_file=" << protocol_to_file 
         << ",  to_stdout=" << protocol_to_stdout
         << ",  make_protocol=" << make_protocol << ")\n";

    if (progressinfo) progressinfo-> text("Merge HDR image...");
    if (progressinfo) progressinfo-> value(0.1);
    
    int         nlayers = pack.size();
    int         dim1    = pack.dim1();
    int         dim2    = pack.dim2();
    const uint8 z_max   = 255;

    const Array1D<double> & logtimes = pack.logtimeVec(); // short name
   
    //  Determine min. threshold z-value, where weight fnct switches to 'not 0'
    uint8  z_threshold_min = 0;
    while ((z_threshold_min < z_max) && (weight[z_threshold_min] == 0.0)) 
      z_threshold_min ++;
    printf("\tz-threshold-min = %d\n", z_threshold_min);
    
#if 1
    //==========================
    //  Beispiel einer 2D Sicht  -  interessant, falls Verschiebungen einzuarbeiten
    //==========================
    //  Alloc (untyped buffer for) HDR image (dim2==W, dim1==H)
    Image  img (dim2, dim1, IMAGE_RGB_F32, "HDR: RGB 32-bit Float");
    //  ...and create a typecasted 2D view on it
    ImgScheme2DView< Rgb<float> >  hdr (img);

    //  largest and least possible HDR value
    Rgb<float> val_min (exp (logX_R[0] - logtimes[nlayers-1]),
                        exp (logX_G[0] - logtimes[nlayers-1]),
                        exp (logX_B[0] - logtimes[nlayers-1]));

    Rgb<float> val_max (exp (logX_R[z_max] - logtimes[0]),
                        exp (logX_G[z_max] - logtimes[0]),
                        exp (logX_B[z_max] - logtimes[0]));

    float max_float = std::numeric_limits<float>::max();
    Rgb<float> h_min ( max_float);  // Init with largest possible value
    Rgb<float> h_max (-max_float);  // Init with least possible value

    double zeit;
    Stopwatch uhr;  uhr.start();

    //  For all pixels
    for (int y=0; y < dim1; y++)
    for (int x=0; x < dim2; x++)
    {
      Rgb<double> sum_w (0.0);          // for summation of weights
      Rgb<double> sum_e (0.0);          // for summation of radiance values
      
      for (int p=0; p < nlayers; p++)
      {
        Rgb<uint8>  z = pack [p] [y] [x];
        Rgb<double>  w (weight[z.r], weight[z.g], weight[z.b]);

        sum_w += w;
//      sum_e += w * Rgb<double>(..., ..., ...)
        sum_e.r += w.r * (logX_R[z.r] - logtimes[p]);
        sum_e.g += w.g * (logX_G[z.g] - logtimes[p]);
        sum_e.b += w.b * (logX_B[z.b] - logtimes[p]);
      }
      
      Rgb<float>&  h = hdr [y] [x];        // short name

      //==========
      //  Compute "h = exp (sum_e / sum_w)", consider cases with sum_w==0
      //
      //  Ob sum_w==0 durch z<z_threshold_min [frueher: z==0] oder z>z_threshold_max 
      //   [frueher: z==z_max] (in allen Bildern) verursacht worden war, wird festgestellt,
      //   indem der Wert in (irgend)einem  Bild angeschaut wird, wir nehmen Bild 0. 
      //   Theoretisch ist sum_w==0 auch in Kombination von z==0 UND z==zmax moeglich, 
      //   aber pathologisch und hier voererst nicht beruecksichtigt.
      // 
      //  WARNUNG:  w!=0 ist nicht ausreichend fuer sicheres exp(sum_e/sum_w)! 
      //  Sicherer waere 
      //    if (sum_w.r != 0)
      //      if (sum_e.r/sum_w.r <= max_exp_arg)  h.r=exp(...)
      //      else h.r=...?
      
#if 1
      //  Kanaele bei Nullwichtung separat auf Grenzwerte gesetzt
      if (sum_w.r != 0.0)
          h.r = exp (sum_e.r / sum_w.r);        // float <- double
      else
      {   if (pack [0][y][x].r < z_threshold_min) // nehmen bspielhaft Bild 0
              h.r = val_min.r;
          else h.r = val_max.r;                 // also z > z_threshold_max
      }
      
      if (sum_w.g != 0.0)
          h.g = exp (sum_e.g / sum_w.g);        // float <- double
      else
      {   if (pack [0][y][x].g < z_threshold_min) // nehmen bspielhaft Bild 0
              h.g = val_min.g;
          else h.g = val_max.g;                 // also z > z_threshold_max
      }

      if (sum_w.b != 0.0)
          h.b = exp (sum_e.b / sum_w.b);        // float <- double
      else
      {   if (pack [0][y][x].b < z_threshold_min) // nehmen bspielhaft Bild 0
              h.b = val_min.b;
          else h.b = val_max.b;                 // also z > z_threshold_max
      }
#else
      //  Probeweise: Sobald ein Kanal nullgewichtet, ganzen RGB auf Grenzwert
      //   setzen. Vermeidet schief abgeschnittene Farben.
      if (sum_w.r == 0.0) 
          if (pack [0][y][x].r < z_threshold_min) // nehmen bspielhaft Bild 0
              h = Rgb<float> (val_min);   
          else                                  // also z > z_threshold_max        
              h = Rgb<float> (val_max);
      else
      if (sum_w.g == 0.0)
          if (pack [0][y][x].g < z_threshold_min) // nehmen bspielhaft Bild 0
              h = Rgb<float> (val_min);   
          else                                  // also z > z_threshold_max
              h = Rgb<float> (val_max);
      else
      if (sum_w.b == 0.0)
          if (pack [0][y][x].b < z_threshold_min) // nehmen bspielhaft Bild 0
              h = Rgb<float> (val_min);   
          else                                  // also z > z_threshold_max
              h = Rgb<float> (val_max);
      else {                                    // also sum_w != 0
          h.r = exp (sum_e.r / sum_w.r);        // float <- double
          h.g = exp (sum_e.g / sum_w.g);        // float <- double
          h.b = exp (sum_e.b / sum_w.b);        // float <- double
      }
#endif      
      //  Die Farbkanaele wurden oben unabhaengig behandelt. Fuer sum_w!=0 ist alles ok,
      //   fuer sum_w.x == 0 aber wurde x-ter Kanal auf Min- bzw. Max-Wert gesetzt.
      //   Falls total sum_w==0, nullte Wichtungsfunktion *alle* Kanaele, Pixel gesetzt 
      //   auf Maximal-"Weiss" bzw. Maximal-"Schwarz". Das ist halbwegs ok, der Pixel kann auch
      //   im Helligkeits-Zoom nicht anders als Schwarz oder Weiss erscheinen. Problematischer,
      //   wenn nur teilw. gesetzt wurde (nur einige sum_w.x==0). Durch das Setzen wurde 
      //   die RGB-Relation geaendert und im Helligkeits-Zoom koennen diese als echte Farben
      //   erscheinen, allerdings verfaelschte, "schief abgeschnittene" Farben.
#if 0
      //===========
      //  BEISPIEL einer effektiven Suche folgender Art:  (aufgehoben)
      //===========
      //  Klaere, ob schlechter (=gesetzter) Pixel, ob dabei vollstaendig oder
      //   teilweise und ob minimaler oder maximaler Wert gesetzt wurde.
      // 
      //   An moeglichen 0/1-Belegungen des Rgb `sum_w gibt es die 2^3=8 Moeglichkeiten
      //     000  100 010 001  110 101 011  111
      //   111 ist regulaer, 000 Totalabschnitt, schiefe Farben geben die 6 Faelle
      //          100 010 001  110 101 011
      enum What {
          OK       = 0,
          PURE_MIN = 1,
          PART_MIN = 2,
          PURE_MAX = 3,
          PART_MAX = 4
      }; 
      What what = OK;
      
      if (sum_w.r == 0)                         // also sum==0**
      {   if (sum_w.g != 0 || sum_w.b != 0)     //   also sum==001, 010, 011
            if (h.r == val_min.r) 
                 what = PART_MIN;
            else what = PART_MAX;
          else                                  //   also sum=000
            if (h.r == val_min.r)
                 what = PURE_MIN;
            else what = PURE_MAX;
      }
      else                                      // also sum==1**
      if (sum_w.g == 0)                         //   also sum==100, 101
      {   if (h.g == val_min.g)
               what = PART_MIN;
          else what = PART_MAX;
      }
      else                                      // also sum=11*  
      if (sum_w.b == 0)                         //   also sum==110
      {   if (h.b == val_min.b)
               what = PART_MIN;
          else what = PART_MAX;
      }
            
      switch (what) 
      {
      case PURE_MIN:
          cout << "pure MIN ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';
          break;
      case PART_MIN:
          cout << "part MIN ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';           
          break;
      case PURE_MAX:
          cout << "pure MIN ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';
          break;
      case PART_MAX:          
          cout << "part MAX ["<<x<<", "<<y<<"] = " << h << "   sum_w="<<sum_w<<'\n';           
          break;
      default: ;  // OK
      }
#endif      
      //  Protokoll-Ausgabe: VOR dem Markieren, da Letzteres HDR-Wert aendert
      //   und danach nur die Markierungswerte noch im Protokoll erschienen.
      if (make_protocol)  
        protocol.out (x,y, sum_w, h, val_min, val_max);
             
      //  Mark "bad" (cropped, zero-weighted, unresolved) pixel
      if (mark_bad_pixel) {
        if (sum_w.r == 0) 
          if (h.r == val_min.r)  h.r = 1.0;     // Kontrast zum Minimum
          else                   h.r = 0.0;     // Kontrast zum Maximum
      
        if (sum_w.g == 0) 
          if (h.g == val_min.g)  h.g = 1.0;     // Kontrast zum Minimum
          else                   h.g = 0.0;     // Kontrast zum Maximum

        if (sum_w.b == 0) 
          if (h.b == val_min.b)  h.b = 1.0;     // Kontrast zum Minimum
          else                   h.b = 0.0;     // Kontrast zum Maximum
      }  
        
      //  MinMax feststellen... (fakultativ)
      if (h.r < h_min.r) h_min.r = h.r; else
      if (h.r > h_max.r) h_max.r = h.r;

      if (h.g < h_min.g) h_min.g = h.g; else
      if (h.g > h_max.g) h_max.g = h.g;

      if (h.b < h_min.b) h_min.b = h.b; else
      if (h.b > h_max.b) h_max.b = h.b;
    
    }  // for all pixels

    //  Weil bei streng fallenden Folgen obige else-Zweige nie erreicht wuerden,
    //   pruefen, ob `h_max geaendert wurde; wenn nicht, ist Maximalwert der 
    //   erste, denn wie gesagt: streng fallend...

    if (h_max.r == -max_float) h_max.r = hdr[0][0].r;
    if (h_max.g == -max_float) h_max.g = hdr[0][0].g;
    if (h_max.b == -max_float) h_max.b = hdr[0][0].b;

#else
    //==========================
    //  Beispiel einer 1D-Sicht  -  spart die eine Multipl. der [][]-Aufrufe.
    //==========================     (Aber noch ohne Markieren und Protokoll!)
    //  Alloc (untyped buffer for) HDR image (dim2==W, dim1==H)
    Image  img (dim2, dim1, IMAGE_RGB_F32, "HDR: RGB 32-bit Float");
    //  ...and create a typecasted 1D view on it
    ImgView1D< Rgb<float> >  hdr(img);

    //  largest and least possible HDR value
    Rgb<float> val_min (exp (logX_R[0] - logtimes[nlayers-1]),
                        exp (logX_G[0] - logtimes[nlayers-1]),
                        exp (logX_B[0] - logtimes[nlayers-1]));

    Rgb<float> val_max (exp (logX_R[z_max] - logtimes[0]),
                        exp (logX_G[z_max] - logtimes[0]),
                        exp (logX_B[z_max] - logtimes[0]));

    float max_float = std::numeric_limits<float>::max();
    Rgb<float> h_min ( max_float);  // Init with largest possible value
    Rgb<float> h_max (-max_float);  // Init with least possible value

    double zeit;
    Stopwatch uhr;  uhr.start();

    //  For all pixels
    for (int i=0; i < hdr.dim(); i++)
    {
      Rgb<double> sum_w (0.0);          // for summation of weights
      Rgb<double> sum_e (0.0);          // for summation of radiance values
      
      for (int p=0; p < nlayers; p++)
      {
        //const Rgb<uint8>&  z = pack [p] . view1d(i);  // short name
        Rgb<uint8>  z = pack [p] . view1d(i);
        Rgb<double>  w (weight[z.r], weight[z.g], weight[z.b]);
        
        sum_w += w;
//      sum_e += w * Rgb<double>(..., ..., ...)
        sum_e.r += w.r * (logX_R[z.r] - logtimes[p]);
        sum_e.g += w.g * (logX_G[z.g] - logtimes[p]);
        sum_e.b += w.b * (logX_B[z.b] - logtimes[p]);
      }
      
      Rgb<float>&  h = hdr [i];            // short name

      //  Compute "h = exp (sum_e / sum_w)", consider cases with sum_w==0
      //
      //  Ob sum_w==0 durch z==0 oder z==zmax (in allen Bildern) verursacht
      //   worden war, wird festgestellt, indem der Wert in (irgend)einem Bild
      //   angeschaut wird, wir nehmen Bild 0. Theoretisch ist sum_w==0 auch in
      //   Kombination von z==0 UND z==zmax moeglich, aber pathologisch und hier
      //   voererst nicht beruecksichtigt.
      //  Das ist fuer n_bad < nlayers sinnvoll.
      
      if (sum_w.r != 0.0)
          h.r = exp (sum_e.r / sum_w.r);  // float <- double
      else
      {   if (pack [0].view1d(i).r == 0)  // nehmen beispielhaft Bild 0
              h.r = val_min.r;            
          else                            // Also z=zmax
              h.r = val_max.r;
      }

      if (sum_w.g != 0.0)
          h.g = exp (sum_e.g / sum_w.g);  // float <- double
      else
      {   if (pack [0].view1d(i).g == 0)  // nehmen beispielhaft Bild 0
              h.g = val_min.g;
          else                            // also z=zmax
              h.g = val_max.g;
      }

      if (sum_w.b != 0.0)
          h.b = exp (sum_e.b / sum_w.b);  // float <- double
      else
      {   if (pack [0].view1d(i).b == 0)  // nehmen beispielhaft Bild 0
              h.b = val_min.b;
          else                            // also z=zmax
              h.b = val_max.b;
      }

      //  MinMax feststellen... (fakultativ)
      if (h.r < h_min.r) h_min.r = h.r; else
      if (h.r > h_max.r) h_max.r = h.r;

      if (h.g < h_min.g) h_min.g = h.g; else
      if (h.g > h_max.g) h_max.g = h.g;

      if (h.b < h_min.b) h_min.b = h.b; else
      if (h.b > h_max.b) h_max.b = h.b;
    
    }  // for all pixels

    //  Weil bei streng fallenden Folgen obige else-Zweige nie erreicht wuerden,
    //   pruefen, ob `h_max geaendert wurde; wenn nicht, ist Maximalwert der
    //   erste, denn wie gesagt: streng fallend.
    
    if (h_max.r == -max_float) h_max.r = hdr[0].r;
    if (h_max.g == -max_float) h_max.g = hdr[0].g;
    if (h_max.b == -max_float) h_max.b = hdr[0].b;
#endif    // 1D-View

    zeit = uhr.stop();
    printf ("\tTime for HDR merging: %f sec\n", zeit);

    cout << "val_min = " << val_min << '\n';
    cout << "val_max = " << val_max << '\n';
    cout << "min = " << h_min << '\n';
    cout << "max = " << h_max << '\n';

    if (progressinfo) progressinfo-> value(1.0);

    return img;
}


/**+*************************************************************************\n
  merge_LogHdr_RGB_U8()  --  description: see ~.hpp and merge_Hdr_RGB_U8().
******************************************************************************/
br::Image    // IMAGE_RGB_F32
merge_LogHdr_RGB_U8 (const PackImgScheme2D_RGB_U8 & pack,
                     const Array1D<double> & logX_R,
                     const Array1D<double> & logX_G,
                     const Array1D<double> & logX_B,
                     const WeightFunc_U8   & weight,
                     bool                  mark_bad_pixel, 
                     ProgressInfo*         progressinfo,
                     bool                  protocol_to_file,
                     bool                  protocol_to_stdout )
{
    cout <<__func__<< "()...\n";
    assert (logX_R.dim() == 256);
    assert (logX_G.dim() == 256);
    assert (logX_B.dim() == 256);
    assert (weight.have_table());
    //assert (weight.dim() == 256);

    //  Evaluate protocol stuff
    const char*       fname = protocol_to_file ? filename_bad_hdr_pixel_ : 0;
    BadPixelProtocol  protocol (fname, protocol_to_stdout);
    bool              make_protocol = protocol.file() || protocol_to_stdout;
    
    cout << "\t(to_file=" << protocol_to_file 
         << ",  to_stdout=" << protocol_to_stdout
         << ",  make_protocol=" << make_protocol << ")\n";
    
    if (progressinfo) progressinfo-> text("Merge logarithmic HDR image...");
    if (progressinfo) progressinfo-> value(0.1);

    //  "Marken-Werte" zur Markierung null-gewichteter Pixel. Sollten Werte
    //   sein, die moeglichst nicht als (logarithm) HDR-Werte auftreten.
    const float  mark_for_Min = -1e30;
    const float  mark_for_Max = -1e31;
    
    int          nlayers = pack.size();
    int          dim1    = pack.dim1();
    int          dim2    = pack.dim2();
    const uint8  z_max   = 255;
    
    const Array1D<double> & logtimes = pack.logtimeVec(); // short name

    //  Determine min. threshold z-value, where weight fnct switches to 'not 0'
    uint8  z_threshold_min = 0;
    while ((z_threshold_min < z_max) && (weight[z_threshold_min] == 0.0)) 
      z_threshold_min ++;
    printf("\tz-threshold-min = %d\n", z_threshold_min);
    
    //==========
    // 2D-Sicht
    //==========
    //  Alloc (untyped buffer for) HDR image (dim2==W, dim1==H)
    Image  img (dim2, dim1, IMAGE_RGB_F32, "HDR: RGB 32-bit Float");
    //  ...and create a typecasted 2D view on it
    ImgScheme2DView< Rgb<float> >  hdr (img);

    //  largest and least possible log. HDR value for this series
    Rgb<float> val_min (logX_R[0] - logtimes[nlayers-1],
                        logX_G[0] - logtimes[nlayers-1],
                        logX_B[0] - logtimes[nlayers-1]);

    Rgb<float> val_max (logX_R[z_max] - logtimes[0],
                        logX_G[z_max] - logtimes[0],
                        logX_B[z_max] - logtimes[0]);

    float max_float = std::numeric_limits<float>::max();
    Rgb<float> h_min ( max_float);      // Init with largest possible value
    Rgb<float> h_max (-max_float);      // Init with smallest possible value

    double zeit;
    Stopwatch uhr;  uhr.start();

    //  For all pixels
    for (int y=0; y < dim1; y++)
    for (int x=0; x < dim2; x++)
    {
      Rgb<double> sum_w (0.0);          // for summation of weights
      Rgb<double> sum_e (0.0);          // for summation of radiance values
      
      for (int p=0; p < nlayers; p++)
      {
        Rgb<uint8>  z = pack [p] [y] [x];
        Rgb<double>  w (weight[z.r], weight[z.g], weight[z.b]);
        
        sum_w += w;
        //  sum_e += w * Rgb<double>(..., ..., ...)
        sum_e.r += w.r * (logX_R[z.r] - logtimes[p]);
        sum_e.g += w.g * (logX_G[z.g] - logtimes[p]);
        sum_e.b += w.b * (logX_B[z.b] - logtimes[p]);
      }
      
      Rgb<float>&  h = hdr [y] [x];     // short name

      //  Compute "h = sum_e / sum_w", consider cases with sum_w==0!
      //   Comment see: merge_Hdr_RGB_U8().

      if (sum_w.r != 0.0)
          h.r = sum_e.r / sum_w.r;              // float <- double
      else 
      {   if (pack [0][y][x].r < z_threshold_min) // nehmen bspielhaft Bild 0
              h.r = val_min.r;
          else                                  // also z > z_threshold_max
              h.r = val_max.r;
      }

      if (sum_w.g != 0.0)
          h.g = sum_e.g / sum_w.g;              // float <- double
      else 
      {   if (pack [0][y][x].g < z_threshold_min) // nehmen bspielhaft Bild 0
              h.g = val_min.g;
          else                                  // also z > z_threshold_max
              h.g = val_max.g;
      }

      if (sum_w.b != 0.0)
          h.b = sum_e.b / sum_w.b;              // float <- double
      else 
      {   if (pack [0][y][x].b < z_threshold_min) // nehmen bspielhaft Bild 0
              h.b = val_min.b;
          else                                  // also z > z_threshold_max
              h.b = val_max.b;
      }

      //  Protokoll-Ausgabe: VOR dem Markieren, da Letzteres HDR-Wert aendert
      //   und danach nur die Markierungswerte noch im Protokoll erschienen.
      //   Allerdings werden hier die *unskalierten* Werte ausgegeben!
      if (make_protocol)  
        protocol.out (x,y, sum_w, h, val_min, val_max);
             
      //  Mark "bad" (cropped, zero-weighted, unresolved) pixel. Weil wir anschliessend
      //   noch skalieren muessen, koennen nicht hier schon die Kontrastwerte gesetzt
      //   werden. Wir setzen stattdessen "Marken-Werte", die wir spaeter wieder 
      //   identifizieren koennen. Zugleich sind -- obligat fuer eine Skalierung --
      //   die Min-Max-Werte zu suchen. Die Marken-Werte sind dabei natuerlich 
      //   auszusparen. Deshalb sind Min-Max-Suche und Markieren hier verschraenkt.
      if (mark_bad_pixel) 
      {
        if (sum_w.r == 0) {
          if (h.r == val_min.r)  h.r = mark_for_Min;
          else                   h.r = mark_for_Max;
        } 
        else {  //  MinMax feststellen
          if (h.r < h_min.r) h_min.r = h.r; else
          if (h.r > h_max.r) h_max.r = h.r;
        }
        
        if (sum_w.g == 0) {
          if (h.g == val_min.g)  h.g = mark_for_Min;
          else                   h.g = mark_for_Max;
        }
        else {  //  MinMax feststellen
          if (h.g < h_min.g) h_min.g = h.g; else
          if (h.g > h_max.g) h_max.g = h.g;
        }

        if (sum_w.b == 0) {
          if (h.b == val_min.b)  h.b = mark_for_Min;
          else                   h.b = mark_for_Max;
        }
        else {  //  MinMax feststellen
          if (h.b < h_min.b) h_min.b = h.b; else
          if (h.b > h_max.b) h_max.b = h.b;
        }
      }  
      else {    //  MinMax feststellen ohne Markieren (einfach)
        if (h.r < h_min.r) h_min.r = h.r; else
        if (h.r > h_max.r) h_max.r = h.r;

        if (h.g < h_min.g) h_min.g = h.g; else
        if (h.g > h_max.g) h_max.g = h.g;

        if (h.b < h_min.b) h_min.b = h.b; else
        if (h.b > h_max.b) h_max.b = h.b;
      }
/*
      if (sum_w.r == 0.0)
      { printf ("pixel(%d,%d),  sum_e.r=%f", x,y, sum_e.r);
        for (int p=0; p < nlayers; p++)
          printf ("  %d", pack [p][y][x].r);
        printf ("\n");
      }
*/
    }  // for all pixels
    
    //  Weil bei streng fallenden Folgen obige else-Zweige nie erreicht wuerden,
    //   pruefen, ob h_max geaendert wurde; wenn nicht, ist Maximalwert der
    //   erste, denn wie gesagt: *streng fallend*.
    //   Falls allerdings "bad pixel" markiert wurden, koennte erster Pixel einen
    //   Markierungswert haben. In diesem Fall ist der erste nichtmarkierte zu
    //   suchen, welcher dann auch der erste in die Min-Max-Suche eingegangene.
    if (mark_bad_pixel) {
      int i;
      if (h_max.r == -max_float) {      // Suche ersten Nichtmarkierten
        i = 0;
        while ((hdr.view1d(i).r == mark_for_Min || 
                hdr.view1d(i).r == mark_for_Max) && i < hdr.size())  i++;
        h_max.r = hdr.view1d(i).r;
      }                
      if (h_max.g == -max_float) {      // Suche ersten Nichtmarkierten
        i = 0;
        while ((hdr.view1d(i).g == mark_for_Min || 
                hdr.view1d(i).g == mark_for_Max) && i < hdr.size())  i++;
        h_max.g = hdr.view1d(i).g;
      }                
      if (h_max.b == -max_float) {      // Suche ersten Nichtmarkierten
        i = 0;
        while ((hdr.view1d(i).b == mark_for_Min || 
                hdr.view1d(i).b == mark_for_Max) && i < hdr.size())  i++;
        h_max.b = hdr.view1d(i).b;
      }                
    }
    else {  // ! mark_bad_pixel
      if (h_max.r == -max_float) h_max.r = hdr[0][0].r;
      if (h_max.g == -max_float) h_max.g = hdr[0][0].g;
      if (h_max.b == -max_float) h_max.b = hdr[0][0].b;
    }

    zeit = uhr.stop();
    printf ("\tTime for Log-HDR merging: %f sec\n", zeit);

    cout << "val_min = " << val_min << '\n';
    cout << "val_max = " << val_max << '\n';
    cout << "min = " << h_min << '\n';
    cout << "max = " << h_max << '\n';

    //  Logarithm domain is (-infty, +infty). Normalize values to [0,1]:
#ifdef SCALE_EACH_CHANNEL_INDEPENDENTLY
    //
    //  Scale each channel independently to [0,1]. Causes color shifts!!
    //   Aber nicht mehr auf aktuellem Stand, Markierung unberuechsichtigt.
    //   Aufbewahrt wegen der schneidig-schoenen Rgb-Schreibweise.
    // 
    Rgb<float> fac = Rgb<float>(1.0) / (h_max - h_min);

    for (int i=0; i < hdr.size(); i++)          // 1D-View 
      hdr.view1d(i) = fac * (hdr.view1d(i) - h_min);
#else
    //
    //  Scale so, that overall-max -> 1, overall-min -> 0:
    //
    float max_all = max (max(h_max.r, h_max.g), max(h_max.g, h_max.b));  
    float min_all = min (min(h_min.r, h_min.g), min(h_min.g, h_min.b));  
    float fac_all = 1.0 / (max_all - min_all);

    cout << "max_all=" << max_all << "  min_all=" << min_all << "  fac_all=" << fac_all << '\n'; 
    
    //  Falls "bad pixel" markiert wurden, diese jetzt raussuchen und auf ihre
    //   Kontrastwerte setzen. Falls nicht markiert wurde, schlicht skalieren.
    if (mark_bad_pixel)
      for (int i=0; i < hdr.size(); i++)          // 1D-View
      {
        Rgb<float>& h = hdr.view1d(i);            // short name
      
        if (h.r == mark_for_Min)       h.r = 1.0;     // Kontrast zum Minimum
        else if (h.r == mark_for_Max)  h.r = 0.0;     // Kontrast zum Minimum
        else                           h.r = fac_all * (h.r - min_all);
        
        if (h.g == mark_for_Min)       h.g = 1.0;     // Kontrast zum Minimum
        else if (h.g == mark_for_Max)  h.g = 0.0;     // Kontrast zum Minimum
        else                           h.g = fac_all * (h.g - min_all);
        
        if (h.b == mark_for_Min)       h.b = 1.0;     // Kontrast zum Minimum
        else if (h.b == mark_for_Max)  h.b = 0.0;     // Kontrast zum Minimum
        else                           h.b = fac_all * (h.b - min_all);
      }
    else
      for (int i=0; i < hdr.size(); i++)          // 1D-View
        hdr.view1d(i) = fac_all * (hdr.view1d(i) - min_all);
#endif
      
    if (progressinfo) progressinfo-> value(1.0);

    return img;
}



/**+*************************************************************************\n
  merge_Hdr_RGB_U16()  --  see: merge_Hdr_RGB_U8()
  
  Provisorisch, um ins momentane HdrManager-Regime zu passen: Wir interpretieren
   die logX-Arrays als Gitterwerte einer tabellierten Responsefunktion zu den
   Stuetzstellen z=256*i+127. Die Spline-y2-Tabellen werden hier erzeugt. Besser
   sollten diese Responsefunktionen als Argument uebergeben werden, was erlaubte,
   die Funktionen ausserhalb zu initialisieren (spline oder tabellieren) und so 
   mehrfach zu verwenden.
******************************************************************************/
br::Image    // IMAGE_RGB_F32
merge_Hdr_RGB_U16 (const br::PackImgScheme2D_RGB_U16 & pack,
                   const TNT::Array1D<double> & logX_R,
                   const TNT::Array1D<double> & logX_G,
                   const TNT::Array1D<double> & logX_B,
                   const WeightFunc_U16       & weight,
                   bool                       mark_bad_pixel, 
                   ProgressInfo*              progressinfo,
                   bool                       protocol_to_file,
                   bool                       protocol_to_stdout )
{
    cout <<__func__<< "(Array1D)...\n";
    
    assert (logX_R.dim() == 256);
    assert (logX_G.dim() == 256);
    assert (logX_B.dim() == 256);
    
    //  Create interpolating response functions from the table values
    ResponseFunc_U16  resp_R (logX_R),
                      resp_G (logX_G),
                      resp_B (logX_B);
    
    return merge_Hdr_RGB_U16 (pack, resp_R, resp_G, resp_B, weight, mark_bad_pixel, 
                              progressinfo, protocol_to_file, protocol_to_stdout); 
}

/**+*************************************************************************\n
  merge_Hdr_RGB_U16()  --  see: merge_Hdr_RGB_U8()
******************************************************************************/
br::Image    // IMAGE_RGB_F32
merge_Hdr_RGB_U16 (const br::PackImgScheme2D_RGB_U16 & pack,
                   const ResponseFunc_U16 & logX_R,
                   const ResponseFunc_U16 & logX_G,
                   const ResponseFunc_U16 & logX_B,
                   const WeightFunc_U16   & weight,
                   bool                     mark_bad_pixel, 
                   ProgressInfo*            progressinfo,
                   bool                     protocol_to_file,
                   bool                     protocol_to_stdout )
{
    cout <<__func__<< "(ResponseFunc)...\n";
    //assert (weight.have_table());

    const char*       fname = protocol_to_file ? filename_bad_hdr_pixel_ : 0;
    BadPixelProtocol  protocol (fname, protocol_to_stdout);
    bool              make_protocol = protocol.file() || protocol_to_stdout;
    cout << "\t(to_file=" << protocol_to_file 
         << ",  to_stdout=" << protocol_to_stdout
         << ",  make_protocol=" << make_protocol << ")\n";
    
    if (progressinfo) progressinfo-> text("Merge HDR image...");
    if (progressinfo) progressinfo-> value(0.1);

    int          nlayers = pack.size();
    int          dim1    = pack.dim1();
    int          dim2    = pack.dim2();
    const uint16 z_max   = 65535;
    
    const Array1D<double> & logtimes = pack.logtimeVec(); // short name

    //  Determine min. threshold z-value, where weight fnct switches to 'not 0'
    uint16  z_threshold_min = 0;
    while ((z_threshold_min < z_max) && (weight(z_threshold_min) == 0.0)) 
      z_threshold_min ++;
    printf("\tz-threshold-min = %d\n", z_threshold_min);
    
    //===========
    // 2D-Sicht
    //===========
    //  Alloc (untyped buffer for) HDR image (dim2==W, dim1==H)
    Image  img (dim2, dim1, IMAGE_RGB_F32, "HDR: RGB 32-bit Float");
    //  ...and create a typecasted 2D-view on it
    ImgScheme2DView< Rgb<float> >  hdr (img);

    //  largest and least possible HDR value
    Rgb<float> val_min (exp (logX_R(0) - logtimes[nlayers-1]),
                        exp (logX_G(0) - logtimes[nlayers-1]),
                        exp (logX_B(0) - logtimes[nlayers-1]));

    Rgb<float> val_max (exp (logX_R(z_max) - logtimes[0]),
                        exp (logX_G(z_max) - logtimes[0]),
                        exp (logX_B(z_max) - logtimes[0]));

    float max_float = std::numeric_limits<float>::max();
    Rgb<float>  h_min ( max_float);  // Init with largest possible value
    Rgb<float>  h_max (-max_float);  // Init with least possible value

    double zeit;
    Stopwatch uhr;  uhr.start();

    for (int y=0; y < dim1; y++)        // Hier genuegte 1D-Sicht
    for (int x=0; x < dim2; x++)
    {
      Rgb<double> sum_w (0.0);          // for summation of weights
      Rgb<double> sum_e (0.0);          // for summation of radiance values
      
      for (int p=0; p < nlayers; p++)
      {
        Rgb<uint16>  z = pack [p] [y] [x];
        Rgb<double>  w;
        if (weight.have_table())  
             w = Rgb<double>(weight[z.r], weight[z.g], weight[z.b]);
        else w = Rgb<double>(weight(z.r), weight(z.g), weight(z.b));

        sum_w += w;
//      sum_e += w * Rgb<double>(..., ..., ...)
        sum_e.r += w.r * (logX_R(z.r) - logtimes[p]);
        sum_e.g += w.g * (logX_G(z.g) - logtimes[p]);
        sum_e.b += w.b * (logX_B(z.b) - logtimes[p]);
      }
      
      Rgb<float>&  h = hdr [y] [x];        // short name

      //  Compute "h = exp (sum_e / sum_w)", consider cases with sum_w==0
      //
      //  Ob sum_w==0 durch z==0 oder z==zmax (in allen Bildern) verursacht
      //   worden war, wird festgestellt, indem der Wert in (irgend)einem Bild
      //   angeschaut wird, wir nehmen Bild 0. Theoretisch ist sum_w==0 auch in
      //   Kombination von z==0 UND z==zmax moeglich, aber pathologisch und hier
      //   vorerst nicht beruecksichtigt. 
      
      if (sum_w.r != 0.0)
          h.r = exp (sum_e.r / sum_w.r);        // float <- double
      else
      {   if (pack [0][y][x].r < z_threshold_min) // nehmen beispielhaft Bild 0
              h.r = val_min.r;            
          else                                  // also z > z_threshold_max
              h.r = val_max.r;
      }

      if (sum_w.g != 0.0)
          h.g = exp (sum_e.g / sum_w.g);        // float <- double
      else
      {   if (pack [0][y][x].g < z_threshold_min) // nehmen beispielhaft Bild 0
              h.g = val_min.g;
          else                                  // also z > z_threshold_max
              h.g = val_max.g;
      }

      if (sum_w.b != 0.0)
          h.b = exp (sum_e.b / sum_w.b);        // float <- double
      else
      {   if (pack [0][y][x].b < z_threshold_min) // nehmen beispielhaft Bild 0
              h.b = val_min.b;
          else                                  // also z > z_threshold_max
              h.b = val_max.b;
      }

      //  Protokoll-Ausgabe: vor dem Markieren, da Letzteres `h aendert und
      //   danach nur die Markierungswerte noch im Protokoll erschienen.
      if (make_protocol)  
        protocol.out (x,y, sum_w, h, val_min, val_max);
             
      //  Mark "bad" (cropped, unresolved) pixel
      if (mark_bad_pixel) {
        if (sum_w.r == 0) 
          if (h.r == val_min.r)  h.r = 1.0;     // Kontrast zum Minimum
          else                   h.r = 0.0;     // Kontrast zum Maximum
      
        if (sum_w.g == 0) 
          if (h.g == val_min.g)  h.g = 1.0;     // Kontrast zum Minimum
          else                   h.g = 0.0;     // Kontrast zum Maximum

        if (sum_w.b == 0) 
          if (h.b == val_min.b)  h.b = 1.0;     // Kontrast zum Minimum
          else                   h.b = 0.0;     // Kontrast zum Maximum
      }  
      
      //  MinMax feststellen... (fakultativ)
      if (h.r < h_min.r) h_min.r = h.r; else
      if (h.r > h_max.r) h_max.r = h.r;

      if (h.g < h_min.g) h_min.g = h.g; else
      if (h.g > h_max.g) h_max.g = h.g;

      if (h.b < h_min.b) h_min.b = h.b; else
      if (h.b > h_max.b) h_max.b = h.b;
    }

    //  Weil bei streng fallenden Folgen obige else-Zweige nie erreicht
    //   wuerden, pruefen, ob h_max geaendert wurde; wenn nicht, ist
    //   Maximalwert der erste, denn wie gesagt: streng fallend...

    if (h_max.r == -max_float) h_max.r = hdr[0][0].r;
    if (h_max.g == -max_float) h_max.g = hdr[0][0].g;
    if (h_max.b == -max_float) h_max.b = hdr[0][0].b;

    zeit = uhr.stop();
    printf ("\tTime for HDR merging: %f sec\n", zeit);

    cout << "val_min = " << val_min << '\n';
    cout << "val_max = " << val_max << '\n';
    cout << "min = " << h_min << '\n';
    cout << "max = " << h_max << '\n';

    if (progressinfo) progressinfo-> value(1.0);

    return img;
}


/**+*************************************************************************\n
  merge_Hdr_RGB_U16 (Array1D)  
   
  Provisorisch... (see: merge_Hdr_RGB_U16(Array1D)).
******************************************************************************/
br::Image    // IMAGE_RGB_F32
merge_LogHdr_RGB_U16 (const br::PackImgScheme2D_RGB_U16 & pack,
                      const TNT::Array1D<double> & logX_R,
                      const TNT::Array1D<double> & logX_G,
                      const TNT::Array1D<double> & logX_B,
                      const WeightFunc_U16       & weight,
                      bool                       mark_bad_pixel, 
                      ProgressInfo*              progressinfo,
                      bool                       protocol_to_file,
                      bool                       protocol_to_stdout )
{
    cout <<__func__<< "(Array1D)...\n";
    
    assert (logX_R.dim() == 256);
    assert (logX_G.dim() == 256);
    assert (logX_B.dim() == 256);
    
    //  Create interpolating response functions from the table values
    ResponseFunc_U16  resp_R (logX_R),
                      resp_G (logX_G),
                      resp_B (logX_B);
    
    return merge_LogHdr_RGB_U16 (pack, resp_R, resp_G, resp_B, weight, mark_bad_pixel, 
                                 progressinfo, protocol_to_file, protocol_to_stdout); 
}

/**+*************************************************************************\n
  merge_LogHdr_RGB_U16 (ResponseFunc_U16)  --  see merge_Hdr_RGB_U16().
******************************************************************************/
br::Image    // IMAGE_RGB_F32
merge_LogHdr_RGB_U16 (const br::PackImgScheme2D_RGB_U16 & pack,
                      const ResponseFunc_U16 & logX_R,
                      const ResponseFunc_U16 & logX_G,
                      const ResponseFunc_U16 & logX_B,
                      const WeightFunc_U16   & weight,
                      bool                     mark_bad_pixel, 
                      ProgressInfo*            progressinfo,
                      bool                     protocol_to_file,
                      bool                     protocol_to_stdout )
{
    cout <<__func__<< "( ResponseFunc_U16 )...\n";
    //assert (weight.have_table());

    //  Evaluate protocol stuff
    const char*       fname = protocol_to_file ? filename_bad_hdr_pixel_ : 0;
    BadPixelProtocol  protocol (fname, protocol_to_stdout);
    bool              make_protocol = protocol.file() || protocol_to_stdout;
    
    cout << "\t(to_file=" << protocol_to_file 
         << ",  to_stdout=" << protocol_to_stdout
         << ",  make_protocol=" << make_protocol << ")\n";
    
    if (progressinfo) progressinfo-> text("Merge logarithmic HDR image...");
    if (progressinfo) progressinfo-> value(0.1);

    //  "Marken-Werte" zur Markierung null-gewichteter Pixel. Sollten Werte
    //   sein, die moeglichst nicht als (logarithm) HDR-Werte auftreten.
    const float  mark_for_Min = -1e30;
    const float  mark_for_Max = -1e31;
    
    int          nlayers = pack.size();
    int          dim1    = pack.dim1();
    int          dim2    = pack.dim2();
    const uint16 z_max   = 65535;
    
    const Array1D<double> & logtimes = pack.logtimeVec(); // short name

    //  Determine min. threshold z-value, where weight fnct switches to 'not 0'
    uint16  z_threshold_min = 0;
    while ((z_threshold_min < z_max) && (weight(z_threshold_min) == 0.0)) 
      z_threshold_min ++;
    printf("\tz-threshold-min = %u\n", z_threshold_min);
    
    //==========
    // 2D-Sicht
    //==========
    //  Alloc (untyped buffer for) HDR image (dim2==W, dim1==H)
    Image  img (dim2, dim1, IMAGE_RGB_F32, "HDR: RGB 32-bit Float");
    //  ...and create a typecasted 2D view on it
    ImgScheme2DView< Rgb<float> >  hdr (img);

    //  largest and least possible log. HDR value for this series
    Rgb<float> val_min (logX_R(0) - logtimes[nlayers-1],
                        logX_G(0) - logtimes[nlayers-1],
                        logX_B(0) - logtimes[nlayers-1]);

    Rgb<float> val_max (logX_R(z_max) - logtimes[0],
                        logX_G(z_max) - logtimes[0],
                        logX_B(z_max) - logtimes[0]);

    float max_float = std::numeric_limits<float>::max();
    Rgb<float> h_min ( max_float);      // Init with largest possible value
    Rgb<float> h_max (-max_float);      // Init with smallest possible value

    double zeit;
    Stopwatch uhr;  uhr.start();

    //  For all pixels
    for (int y=0; y < dim1; y++)
    for (int x=0; x < dim2; x++)
    {
      Rgb<double> sum_w (0.0);          // for summation of weights
      Rgb<double> sum_e (0.0);          // for summation of radiance values
      
      for (int p=0; p < nlayers; p++)
      {
        Rgb<uint16>  z = pack [p] [y] [x];
        Rgb<double>  w;  
        if (weight.have_table())  
             w = Rgb<double>(weight[z.r], weight[z.g], weight[z.b]);
        else w = Rgb<double>(weight(z.r), weight(z.g), weight(z.b)); 
          
        sum_w += w;
        //  sum_e += w * Rgb<double>(..., ..., ...)
        sum_e.r += w.r * (logX_R(z.r) - logtimes[p]);
        sum_e.g += w.g * (logX_G(z.g) - logtimes[p]);
        sum_e.b += w.b * (logX_B(z.b) - logtimes[p]);
      }
      
      Rgb<float>&  h = hdr [y] [x];     // short name

      //  Compute "h = sum_e / sum_w", consider cases with sum_w==0!
      //   Comment see merge_Hdr_RGB_U8().

      if (sum_w.r != 0.0)
          h.r = sum_e.r / sum_w.r;              // float <- double
      else 
      {   if (pack [0][y][x].r < z_threshold_min) // nehmen bspielhaft Bild 0
              h.r = val_min.r;
          else                                  // also z > z_threshold_max
              h.r = val_max.r;
      }

      if (sum_w.g != 0.0)
          h.g = sum_e.g / sum_w.g;              // float <- double
      else 
      {   if (pack [0][y][x].g < z_threshold_min) // nehmen bspielhaft Bild 0
              h.g = val_min.g;
          else                                  // also z > z_threshold_max
              h.g = val_max.g;
      }

      if (sum_w.b != 0.0)
          h.b = sum_e.b / sum_w.b;              // float <- double
      else 
      {   if (pack [0][y][x].b < z_threshold_min) // nehmen bspielhaft Bild 0
              h.b = val_min.b;
          else                                  // also z > z_threshold_max
              h.b = val_max.b;
      }

      //  Protokoll-Ausgabe: VOR dem Markieren, da Letzteres HDR-Wert aendert
      //   und danach nur die Markierungswerte noch im Protokoll erschienen.
      //   Allerdings werden hier die *unskalierten* Werte ausgegeben!
      if (make_protocol)  
        protocol.out (x,y, sum_w, h, val_min, val_max);
             
      //  Mark "bad" (cropped, zero-weighted, unresolved) pixel. Weil wir anschliessend
      //   noch skalieren muessen, koennen nicht hier schon die Kontrastwerte gesetzt
      //   werden. Wir setzen stattdessen "Marken-Werte", die wir spaeter wieder 
      //   identifizieren koennen. Zugleich sind -- obligat fuer eine Skalierung --
      //   die Min-Max-Werte zu suchen. Die Marken-Werte sind dabei natuerlich 
      //   auszusparen. Deshalb sind Min-Max-Suche und Markierung hier verschraenkt.
      if (mark_bad_pixel) 
      {
        if (sum_w.r == 0) {
          if (h.r == val_min.r)  h.r = mark_for_Min;
          else                   h.r = mark_for_Max;
        } 
        else {  //  MinMax feststellen
          if (h.r < h_min.r) h_min.r = h.r; else
          if (h.r > h_max.r) h_max.r = h.r;
        }
        
        if (sum_w.g == 0) {
          if (h.g == val_min.g)  h.g = mark_for_Min;
          else                   h.g = mark_for_Max;
        }
        else {  //  MinMax feststellen
          if (h.g < h_min.g) h_min.g = h.g; else
          if (h.g > h_max.g) h_max.g = h.g;
        }

        if (sum_w.b == 0) {
          if (h.b == val_min.b)  h.b = mark_for_Min;
          else                   h.b = mark_for_Max;
        }
        else {  //  MinMax feststellen
          if (h.b < h_min.b) h_min.b = h.b; else
          if (h.b > h_max.b) h_max.b = h.b;
        }
      }  
      else {    //  MinMax feststellen ohne Markieren (einfach)
        if (h.r < h_min.r) h_min.r = h.r; else
        if (h.r > h_max.r) h_max.r = h.r;

        if (h.g < h_min.g) h_min.g = h.g; else
        if (h.g > h_max.g) h_max.g = h.g;

        if (h.b < h_min.b) h_min.b = h.b; else
        if (h.b > h_max.b) h_max.b = h.b;
      }
    
    }  // for all pixels
    
    //  Weil bei streng fallenden Folgen obige else-Zweige nie erreicht wuerden,
    //   pruefen, ob h_max geaendert wurde; wenn nicht, ist Maximalwert der
    //   erste, denn wie gesagt: *streng fallend*.
    //   Falls allerdings "bad pixel" markiert wurden, koennte erste Pixel einen
    //   Markierungswert haben. In diesem Fall ist der erste nichtmarkierte zu
    //   suchen, welcher dann auch der erste in die Min-Max-Suche eingegangene.
    if (mark_bad_pixel) {
      int i;
      if (h_max.r == -max_float) {      // Suche ersten Nichtmarkierten
        i = 0;
        while ((hdr.view1d(i).r == mark_for_Min || 
                hdr.view1d(i).r == mark_for_Max) && i < hdr.size())  i++;
        h_max.r = hdr.view1d(i).r;
      }                
      if (h_max.g == -max_float) {      // Suche ersten Nichtmarkierten
        i = 0;
        while ((hdr.view1d(i).g == mark_for_Min || 
                hdr.view1d(i).g == mark_for_Max) && i < hdr.size())  i++;
        h_max.g = hdr.view1d(i).g;
      }                
      if (h_max.b == -max_float) {      // Suche ersten Nichtmarkierten
        i = 0;
        while ((hdr.view1d(i).b == mark_for_Min || 
                hdr.view1d(i).b == mark_for_Max) && i < hdr.size())  i++;
        h_max.b = hdr.view1d(i).b;
      }                
    }
    else {  
      if (h_max.r == -max_float) h_max.r = hdr[0][0].r;
      if (h_max.g == -max_float) h_max.g = hdr[0][0].g;
      if (h_max.b == -max_float) h_max.b = hdr[0][0].b;
    }

    zeit = uhr.stop();
    printf ("\tTime for Log-HDR merging: %f sec\n", zeit);

    cout << "val_min = " << val_min << '\n';
    cout << "val_max = " << val_max << '\n';
    cout << "min = " << h_min << '\n';
    cout << "max = " << h_max << '\n';

    //  Logarithm domain is (-infty, +infty). Normalize values to [0,1]:
    //   Scale so, that overall-max -> 1, overall-min -> 0:
    float max_all = max (max(h_max.r, h_max.g), max(h_max.g, h_max.b));  
    float min_all = min (min(h_min.r, h_min.g), min(h_min.g, h_min.b));  
    float fac_all = 1.0 / (max_all - min_all);

    cout << "max_all=" << max_all << "  min_all=" << min_all << "  fac_all=" << fac_all << '\n'; 
    
    //  Falls "bad pixel" markiert wurden, diese jetzt raussuchen und auf ihre
    //   Kontrastwerte setzen. Falls nicht markiert wurde, schlicht skalieren.
    if (mark_bad_pixel)
      for (int i=0; i < hdr.size(); i++)          // 1D-View
      {
        Rgb<float>& h = hdr.view1d(i);            // short name
      
        if (h.r == mark_for_Min)       h.r = 1.0;     // Kontrast zum Minimum
        else if (h.r == mark_for_Max)  h.r = 0.0;     // Kontrast zum Minimum
        else                           h.r = fac_all * (h.r - min_all);
        
        if (h.g == mark_for_Min)       h.g = 1.0;     // Kontrast zum Minimum
        else if (h.g == mark_for_Max)  h.g = 0.0;     // Kontrast zum Minimum
        else                           h.g = fac_all * (h.g - min_all);
        
        if (h.b == mark_for_Min)       h.b = 1.0;     // Kontrast zum Minimum
        else if (h.b == mark_for_Max)  h.b = 0.0;     // Kontrast zum Minimum
        else                           h.b = fac_all * (h.b - min_all);
      }
    else
      for (int i=0; i < hdr.size(); i++)          // 1D-View
        hdr.view1d(i) = fac_all * (hdr.view1d(i) - min_all);
      
    if (progressinfo) progressinfo-> value(1.0);
    
    return img;
}


#if 0
/**+*************************************************************************\n
  merge_LogHdr_RGB_U16()
  
  Einfach gemacht: Berechnen erst normales HDR und logarithmieren dann. So 
   natuerlich langsamer statt schneller.
   
  @NOTE  Das derzeitige schlichte Logarithmieren setzt HDR-Werte <> 0 voraus. 
   Wenn das nicht mehr erfuellt, waere fuer hdr==0 ein kleiner Wert zu setzen
   (in der Groessenordnung des kleinsten log-Werte fuer hdr!=0), um anschliessend
   sinnvoll auf [0,1] zu skalieren. "-oo" waere natuerlich sinnlos. Praktisch
   koennte man hdr=limits<float>.min() setzen, aber aus Min-Max-Suche aussparen,
   und bei Skalierung die limit-Werte herausfischen und sinnvoll setzen.
******************************************************************************/
br::Image    // IMAGE_RGB_F32
merge_LogHdr_RGB_U16 (const br::PackImgScheme2D_RGB_U16 & pack,
                      const TNT::Array1D<double> & logX_R,
                      const TNT::Array1D<double> & logX_G,
                      const TNT::Array1D<double> & logX_B,
                      const WeightFunc_U16       & weight,
                      bool                       mark_bad_pixel,
                      ProgressInfo*              progressinfo,
                      bool                       protocol_to_file,
                      bool                       protocol_to_stdout )
{
    cout <<__func__<< "(Array1D)...\n";
    
    //  Compute ordinary HDR
    Image  img = merge_Hdr_RGB_U16 (pack, logX_R, logX_G, logX_B, weight, mark_bad_pixel,
                                    progressinfo, protocol_to_file, protocol_to_stdout);
    img.report();
    if (img.is_null()) return img;
    
    //  Typecasted 1D view on image data
    ImgView1D< Rgb<float> >  hdr (img);
    
    //  Logarithm the first value
    hdr[0] = Rgb<float> (log(hdr[0].r), log(hdr[0].g), log(hdr[0].b));
    //  and init min-max-values with it
    Rgb<float> h_min = hdr[0];
    Rgb<float> h_max = hdr[0];
    
    //  For the rest of the pixels
    for (int i=1; i < hdr.dim(); i++)
    {
      Rgb<float>&  h = hdr[i];          // short name
      h.r = log (h.r);
      h.g = log (h.g);
      h.b = log (h.b);
      
      //  MinMax feststellen
      if (h.r < h_min.r) h_min.r = h.r; else
      if (h.r > h_max.r) h_max.r = h.r;

      if (h.g < h_min.g) h_min.g = h.g; else
      if (h.g > h_max.g) h_max.g = h.g;

      if (h.b < h_min.b) h_min.b = h.b; else
      if (h.b > h_max.b) h_max.b = h.b;
    }
    cout << "min = " << h_min << '\n';
    cout << "max = " << h_max << '\n';

    //  Scale so, that overall-max -> 1, overall-min -> 0:
    float max_all = max (max(h_max.r, h_max.g), max(h_max.g, h_max.b));  
    float min_all = min (min(h_min.r, h_min.g), min(h_min.g, h_min.b));  
    float fac_all = 1.0 / (max_all - min_all);

    cout << "max_all=" << max_all << "  min_all=" << min_all << "  fac_all=" << fac_all << '\n'; 
      
    for (int i=0; i < hdr.dim(); i++)
      hdr[i] = fac_all * (hdr[i] - min_all);
    
    return img;
}                         
#endif


}  // namespace "br"

// END OF FILE
